@@ -55,6 +55,7 @@ from .types import (
     ConstrainedInt,
     ConstrainedList,
     ConstrainedSet,
+    ConstrainedStr,
     SecretBytes,
     SecretStr,
     StrictBytes,
@@ -226,7 +227,7 @@ def field_schema(
     model_name_map: Dict[TypeModelOrEnum, str],
     ref_prefix: Optional[str] = None,
     ref_template: str = default_ref_template,
-    known_models: TypeModelSet = None,
+    known_models: Optional[TypeModelSet] = None,
 ) -> Tuple[Dict[str, Any], Dict[str, Any], Set[str]]:
     """
     Process a Pydantic field and return a tuple with a JSON Schema for it as the first item.
@@ -346,7 +347,7 @@ def get_model_name_map(unique_models: TypeModelSet) -> Dict[TypeModelOrEnum, str
     return {v: k for k, v in name_model_map.items()}
 
 
-def get_flat_models_from_model(model: Type['BaseModel'], known_models: TypeModelSet = None) -> TypeModelSet:
+def get_flat_models_from_model(model: Type['BaseModel'], known_models: Optional[TypeModelSet] = None) -> TypeModelSet:
     """
     Take a single ``model`` and generate a set with itself and all the sub-models in the tree. I.e. if you pass
     model ``Foo`` (subclass of Pydantic ``BaseModel``) as ``model``, and it has a field of type ``Bar`` (also
@@ -368,7 +369,7 @@ def get_flat_models_from_model(model: Type['BaseModel'], known_models: TypeModel
 
 def get_flat_models_from_field(field: ModelField, known_models: TypeModelSet) -> TypeModelSet:
     """
-    Take a single Pydantic ``ModelField`` (from a model) that could have been declared as a sublcass of BaseModel
+    Take a single Pydantic ``ModelField`` (from a model) that could have been declared as a subclass of BaseModel
     (so, it could be a submodel), and generate a set with its model and all the sub-models in the tree.
     I.e. if you pass a field that was declared to be of type ``Foo`` (subclass of BaseModel) as ``field``, and that
     model ``Foo`` has a field of type ``Bar`` (also subclass of ``BaseModel``) and that model ``Bar`` has a field of
@@ -490,7 +491,7 @@ def field_type_schema(
         if regex:
             # Dict keys have a regex pattern
             # items_schema might be a schema or empty dict, add it either way
-            f_schema['patternProperties'] = {regex.pattern: items_schema}
+            f_schema['patternProperties'] = {ConstrainedStr._get_pattern(regex): items_schema}
         if items_schema:
             # The dict values are not simply Any, so they need a schema
             f_schema['additionalProperties'] = items_schema
@@ -555,7 +556,7 @@ def model_process_schema(
     model_name_map: Dict[TypeModelOrEnum, str],
     ref_prefix: Optional[str] = None,
     ref_template: str = default_ref_template,
-    known_models: TypeModelSet = None,
+    known_models: Optional[TypeModelSet] = None,
     field: Optional[ModelField] = None,
 ) -> Tuple[Dict[str, Any], Dict[str, Any], Set[str]]:
     """
@@ -717,6 +718,8 @@ def field_singleton_sub_fields_schema(
             discriminator_models_refs: Dict[str, Union[str, Dict[str, Any]]] = {}
 
             for discriminator_value, sub_field in field.sub_fields_mapping.items():
+                if isinstance(discriminator_value, Enum):
+                    discriminator_value = str(discriminator_value.value)
                 # sub_field is either a `BaseModel` or directly an `Annotated` `Union` of many
                 if is_union(get_origin(sub_field.type_)):
                     sub_models = get_sub_types(sub_field.type_)
@@ -867,7 +870,7 @@ def field_singleton_schema(  # noqa: C901 (ignore complexity)
         f_schema['const'] = field.default
 
     if is_literal_type(field_type):
-        values = all_literal_values(field_type)
+        values = tuple(x.value if isinstance(x, Enum) else x for x in all_literal_values(field_type))
 
         if len({v.__class__ for v in values}) > 1:
             return field_schema(
