@@ -5,6 +5,7 @@ from datetime import datetime
 from typing import List, Optional
 
 from pydantic import BaseModel, NoneStr
+from pydantic.types import Json
 
 
 class Model(BaseModel):
@@ -13,8 +14,10 @@ class Model(BaseModel):
     last_name: NoneStr = None
     signup_ts: Optional[datetime] = None
     list_of_ints: List[int]
+    json_list_of_ints: Json[List[int]]
 
 
 m = Model(age=42, list_of_ints=[1, '2', b'3'])
 
 print(m.age + 'not integer')
+m.json_list_of_ints[0] + 'not integer'
