@@ -1,16 +1,17 @@
 """
-Stolen from https://github.com/django/django/blob/master/tests/utils_tests/test_dateparse.py at
+Stolen from https://github.com/django/django/blob/main/tests/utils_tests/test_dateparse.py at
 9718fa2e8abe430c3526a9278dd976443d4ae3c6
 
 Changed to:
 * use standard pytest layout
 * parametrize tests
 """
+import re
 from datetime import date, datetime, time, timedelta, timezone
 
 import pytest
 
-from pydantic import BaseModel, ValidationError, errors
+from pydantic import BaseModel, ValidationError, condate, errors
 from pydantic.datetime_parse import parse_date, parse_datetime, parse_duration, parse_time
 
 
@@ -175,6 +176,7 @@ def test_parse_python_format(delta):
         ('30', timedelta(seconds=30)),
         (30, timedelta(seconds=30)),
         (30.1, timedelta(seconds=30, milliseconds=100)),
+        (9.9e-05, timedelta(microseconds=99)),
         # minutes seconds
         ('15:30', timedelta(minutes=15, seconds=30)),
         ('5:30', timedelta(minutes=5, seconds=30)),
@@ -289,3 +291,29 @@ def test_nan():
             'type': 'value_error',
         },
     ]
+
+
+@pytest.mark.parametrize(
+    'constraint,msg,ok_value,error_value',
+    [
+        ('gt', 'greater than', date(2020, 1, 2), date(2019, 12, 31)),
+        ('gt', 'greater than', date(2020, 1, 2), date(2020, 1, 1)),
+        ('ge', 'greater than or equal to', date(2020, 1, 2), date(2019, 12, 31)),
+        ('ge', 'greater than or equal to', date(2020, 1, 1), date(2019, 12, 31)),
+        ('lt', 'less than', date(2019, 12, 31), date(2020, 1, 2)),
+        ('lt', 'less than', date(2019, 12, 31), date(2020, 1, 1)),
+        ('le', 'less than or equal to', date(2019, 12, 31), date(2020, 1, 2)),
+        ('le', 'less than or equal to', date(2020, 1, 1), date(2020, 1, 2)),
+    ],
+)
+def test_date_constraints(constraint, msg, ok_value, error_value):
+    class Model(BaseModel):
+        a: condate(**{constraint: date(2020, 1, 1)})
+
+    assert Model(a=ok_value).dict() == {'a': ok_value}
+
+    match = re.escape(
+        f'ensure this value is {msg} 2020-01-01 ' f'(type=value_error.number.not_{constraint}; limit_value=2020-01-01)'
+    )
+    with pytest.raises(ValidationError, match=match):
+        Model(a=error_value)
